@@ -1,8 +1,7 @@
 /*
 Copyright(c) 2002-2005 Anatoliy Kuznetsov(anatoliy_kuznetsov at yahoo.com)
 
-PermiFor more information please visit:  http://bmagic.sourceforge.net
-ssion is hereby granted, free of charge, to any person 
+Permission is hereby granted, free of charge, to any person 
 obtaining a copy of this software and associated documentation 
 files (the "Software"), to deal in the Software without restriction, 
 including without limitation the rights to use, copy, modify, merge, 
@@ -38,6 +37,12 @@ For more information please visit:  http://bmagic.sourceforge.net
 # include <iterator>
 #endif
 
+#ifdef _MSC_VER
+#pragma warning( disable : 4311 4312)
+#endif
+
+
+
 #include "bmconst.h"
 #include "bmdef.h"
 
@@ -113,26 +118,9 @@ public:
     /** Type used to count bits in the bit vector */
     typedef bm::id_t                   size_type; 
 
-    /*!
-       @brief Structure with statistical information about bitset's memory 
-              allocation details. 
-       @ingroup bvector
-    */
-    struct statistics
-    {
-        /// Number of bit blocks.
-        unsigned bit_blocks; 
-        /// Number of GAP blocks.
-        unsigned gap_blocks;  
-        /// Estimated maximum of memory required for serialization.
-        unsigned max_serialize_mem;
-        /// Memory used by bitvector including temp and service blocks
-        unsigned  memory_used;
-        /// Array of all GAP block lengths in the bvector.
-        gap_word_t   gap_length[bm::set_total_blocks];
-        /// GAP lengths used by bvector
-        gap_word_t  gap_levels[bm::gap_levels];
-    };
+    /** Statistical information about bitset's memory allocation details. */
+    struct statistics : public bv_statistics
+    {};
 
     /**
         @brief Class reference implements an object for bit assignment.
@@ -181,7 +169,7 @@ public:
         /*! Bitwise AND. Performs operation: bit = bit AND value */
         const reference& operator&=(bool value) const
         {
-            bv_.set(position_, value);
+            bv_.set_bit_and(position_, value);
             return *this;
         }
 
@@ -236,7 +224,7 @@ public:
     {
     friend class bvector;
     public:
-        iterator_base() : block_(0) {}
+        iterator_base() : bv_(0), position_(bm::id_max), block_(0) {}
 
         bool operator==(const iterator_base& it) const
         {
@@ -961,6 +949,32 @@ public:
         return set_bit_no_check(n, val);
     }
 
+    /*!
+       \brief Sets bit n using bit AND with the provided value.
+       \param n - index of the bit to be set. 
+       \param val - new bit value
+       \return  TRUE if bit was changed
+    */
+    bool set_bit_and(bm::id_t n, bool val = true)
+    {
+        BM_ASSERT(n < size_);
+        return and_bit_no_check(n, val);
+    }
+
+    /*!
+       \brief Sets bit n only if current value is equal to the condition
+       \param n - index of the bit to be set. 
+       \param val - new bit value
+       \param condition - expected current value
+       \return TRUE if bit was changed
+    */
+    bool set_bit_conditional(bm::id_t n, bool val, bool condition)
+    {
+        BM_ASSERT(n < size_);
+        if (val == condition) return false;
+        return set_bit_conditional_impl(n, val, condition);
+    }
+
 
     /*!
         \brief Sets bit n if val is true, clears bit n if val is false
@@ -1082,7 +1096,7 @@ public:
     {
         bm::word_t*** blk_root = blockman_.get_rootblock();
         typename blocks_manager_type::block_count_arr_func func(blockman_, &(arr[0]));
-        for_each_nzblock(blk_root, blockman_.top_block_size(), 
+        for_each_nzblock(blk_root, blockman_.effective_top_block_size(), 
                                    bm::set_array_size, func);
         return func.last_block();
     }
@@ -1153,10 +1167,13 @@ public:
     #endif
         
         word_t*** blk_root = blockman_.get_rootblock();
-        if (!blk_root) return false;
+        if (!blk_root) 
+            return false;
         typename blocks_manager_type::block_any_func func(blockman_);
-        return for_each_nzblock_if(blk_root, blockman_.top_block_size(),
-                                             bm::set_array_size, func);
+        return for_each_nzblock_if(blk_root, 
+                                   blockman_.effective_top_block_size(),
+                                   bm::set_array_size, 
+                                   func);
     }
 
     /*!
@@ -1206,7 +1223,7 @@ public:
        \fn bm::id_t bvector::get_first() const
        \brief Gets number of first bit which is ON.
        \return Index of the first 1 bit.
-       \sa get_next
+       \sa get_next, extract_next
     */
     bm::id_t get_first() const { return check_or_next(0); }
 
@@ -1215,13 +1232,20 @@ public:
        \brief Finds the number of the next bit ON.
        \param prev - Index of the previously found bit. 
        \return Index of the next bit which is ON or 0 if not found.
-       \sa get_first
+       \sa get_first, extract_next
     */
     bm::id_t get_next(bm::id_t prev) const
     {
         return (++prev == bm::id_max) ? 0 : check_or_next(prev);
     }
 
+    /*!
+       \fn bm::id_t bvector::extract_next(bm::id_t prev)
+       \brief Finds the number of the next bit ON and sets it to 0.
+       \param prev - Index of the previously found bit. 
+       \return Index of the next bit which is ON or 0 if not found.
+       \sa get_first, get_next, 
+    */
     bm::id_t extract_next(bm::id_t prev)
     {
         return (++prev == bm::id_max) ? 0 : check_or_next_extract(prev);
@@ -1328,10 +1352,14 @@ public:
        with a regular structure, frees some memory. This function is recommended
        after a bulk modification of the bitvector using set_bit, clear_bit or
        logical operations.
+
+       Optionally function can calculate vector post optimization statistics
        
        @sa optmode, optimize_gap_size
     */
-    void optimize(bm::word_t* temp_block=0, optmode opt_mode = opt_compress);
+    void optimize(bm::word_t* temp_block = 0, 
+                  optmode opt_mode       = opt_compress,
+                  statistics* stat       = 0);
 
     /*!
        \brief Optimize sizes of GAP blocks
@@ -1418,6 +1446,9 @@ public:
         return blockman_.get_block(nb); 
     }
     
+    void combine_operation(const bm::bvector<Alloc, MS>& bvect, 
+                            bm::operation                opcode);
+    
 private:
 
     bm::id_t check_or_next(bm::id_t prev) const;
@@ -1428,13 +1459,17 @@ private:
     bm::id_t check_or_next_extract(bm::id_t prev);
 
     /**
-        \brief Set spacified bit without checking preconditions (size, etc)
+        \brief Set specified bit without checking preconditions (size, etc)
     */
     bool set_bit_no_check(bm::id_t n, bool val);
 
+    /**
+        \brief AND specified bit without checking preconditions (size, etc)
+    */
+    bool and_bit_no_check(bm::id_t n, bool val);
+
+    bool set_bit_conditional_impl(bm::id_t n, bool val, bool condition);
 
-    void combine_operation(const bm::bvector<Alloc, MS>& bvect, 
-                            bm::operation                opcode);
 
     void combine_operation_with_block(unsigned nb,
                                       unsigned gap,
@@ -1606,8 +1641,13 @@ bm::id_t bvector<Alloc, MS>::count() const
     if (count_is_valid_) return count_;
 #endif
     word_t*** blk_root = blockman_.get_rootblock();
+    if (!blk_root) 
+    {
+        BMCOUNT_SET(0);
+        return 0;
+    }    
     typename blocks_manager_type::block_count_func func(blockman_);
-    for_each_nzblock(blk_root, blockman_.top_block_size(), 
+    for_each_nzblock(blk_root, blockman_.effective_top_block_size(), 
                                 bm::set_array_size, func);
 
     BMCOUNT_SET(func.count());
@@ -1787,7 +1827,9 @@ bool bvector<Alloc, MS>::get_bit(bm::id_t n) const
 // -----------------------------------------------------------------------
 
 template<typename Alloc, typename MS> 
-void bvector<Alloc, MS>::optimize(bm::word_t* temp_block, optmode opt_mode)
+void bvector<Alloc, MS>::optimize(bm::word_t* temp_block, 
+                                  optmode     opt_mode,
+                                  statistics* stat)
 {
     word_t*** blk_root = blockman_.blocks_root();
 
@@ -1797,9 +1839,31 @@ void bvector<Alloc, MS>::optimize(bm::word_t* temp_block, optmode opt_mode)
     typename 
         blocks_manager_type::block_opt_func  opt_func(blockman_, 
                                                 temp_block, 
-                                                (int)opt_mode);
-    for_each_nzblock(blk_root, blockman_.top_block_size(),
-                                bm::set_array_size, opt_func);
+                                                (int)opt_mode,
+                                                stat);
+    if (stat)
+    {
+        stat->bit_blocks = stat->gap_blocks 
+                         = stat->max_serialize_mem 
+                         = stat->memory_used 
+                         = 0;
+        ::memcpy(stat->gap_levels, 
+                blockman_.glen(), sizeof(gap_word_t) * bm::gap_levels);
+        stat->max_serialize_mem = sizeof(id_t) * 4;
+
+    }
+
+    for_each_nzblock(blk_root, blockman_.effective_top_block_size(),
+                               bm::set_array_size, opt_func);
+
+    if (stat)
+    {
+        unsigned safe_inc = stat->max_serialize_mem / 10; // 10% increment
+        if (!safe_inc) safe_inc = 256;
+        stat->max_serialize_mem += safe_inc;
+        stat->memory_used += sizeof(*this) - sizeof(blockman_);
+        stat->memory_used += blockman_.mem_used();
+    }
 }
 
 // -----------------------------------------------------------------------
@@ -1847,8 +1911,8 @@ int bvector<Alloc, MS>::compare(const bvector<Alloc, MS>& bvect) const
     int res;
     unsigned bn = 0;
 
-    unsigned top_blocks = blockman_.top_block_size();
-    unsigned bvect_top_blocks = bvect.blockman_.top_block_size();
+    unsigned top_blocks = blockman_.effective_top_block_size();
+    unsigned bvect_top_blocks = bvect.blockman_.effective_top_block_size();
 
     if (bvect_top_blocks > top_blocks) top_blocks = bvect_top_blocks;
 
@@ -1990,8 +2054,9 @@ void bvector<Alloc, MS>::calc_stat(typename bvector<Alloc, MS>::statistics* st)
 
     unsigned block_idx = 0;
 
+    unsigned top_size = blockman_.effective_top_block_size();
     // Walk the blocks, calculate statistics.
-    for (unsigned i = 0; i < blockman_.top_block_size(); ++i)
+    for (unsigned i = 0; i < top_size; ++i)
     {
         const bm::word_t* const* blk_blk = blockman_.get_topblock(i);
 
@@ -2042,8 +2107,9 @@ void bvector<Alloc, MS>::calc_stat(typename bvector<Alloc, MS>::statistics* st)
         }
     }  
 
-
-    st->max_serialize_mem += st->max_serialize_mem / 10; // 10% increment
+    unsigned safe_inc = st->max_serialize_mem / 10; // 10% increment
+    if (!safe_inc) safe_inc = 256;
+    st->max_serialize_mem += safe_inc;
 
     // Calc size of different odd and temporary things.
 
@@ -2126,6 +2192,159 @@ bool bvector<Alloc, MS>::set_bit_no_check(bm::id_t n, bool val)
     return false;
 }
 
+// -----------------------------------------------------------------------
+
+template<class Alloc, class MS> 
+bool bvector<Alloc, MS>::set_bit_conditional_impl(bm::id_t n, 
+                                                  bool     val, 
+                                                  bool     condition)
+{
+    // calculate logical block number
+    unsigned nblock = unsigned(n >>  bm::set_block_shift); 
+
+    int block_type;
+    bm::word_t* blk = 
+        blockman_.check_allocate_block(nblock, 
+                                       val,
+                                       get_new_blocks_strat(), 
+                                       &block_type);
+    if (!blk) 
+        return false;
+
+    // calculate word number in block and bit
+    unsigned nbit   = unsigned(n & bm::set_block_mask); 
+
+    if (block_type == 1) // gap
+    {
+        bm::gap_word_t* gap_blk = BMGAP_PTR(blk);
+        bool old_val = (gap_test(gap_blk, nbit) != 0);
+
+        if (old_val != condition) 
+        {
+            return false;
+        }
+
+        if (val != old_val)
+        {
+            unsigned is_set;
+            unsigned new_block_len = 
+                gap_set_value(val, gap_blk, nbit, &is_set);
+            BM_ASSERT(is_set);
+            BMCOUNT_ADJ(val)
+
+            unsigned threshold = 
+                bm::gap_limit(gap_blk, blockman_.glen());
+            if (new_block_len > threshold) 
+            {
+                extend_gap_block(nblock, gap_blk);
+            }
+            return true;
+        }
+    }
+    else  // bit block
+    {
+        unsigned nword  = unsigned(nbit >> bm::set_word_shift); 
+        nbit &= bm::set_word_mask;
+
+        bm::word_t* word = blk + nword;
+        bm::word_t  mask = (((bm::word_t)1) << nbit);
+        bool is_set = ((*word) & mask) != 0;
+
+        if (is_set != condition)
+        {
+            return false;
+        }
+        if (is_set != val)    // need to change bit
+        {
+            if (val)          // set bit
+            {
+                *word |= mask;
+                BMCOUNT_INC;
+            }
+            else               // clear bit
+            {
+                *word &= ~mask;
+                BMCOUNT_DEC;
+            }
+            return true;
+        }
+    }
+    return false;
+
+}
+
+// -----------------------------------------------------------------------
+
+
+template<class Alloc, class MS> 
+bool bvector<Alloc, MS>::and_bit_no_check(bm::id_t n, bool val)
+{
+    // calculate logical block number
+    unsigned nblock = unsigned(n >>  bm::set_block_shift); 
+
+    int block_type;
+    bm::word_t* blk = 
+        blockman_.check_allocate_block(nblock, 
+                                       val,
+                                       get_new_blocks_strat(), 
+                                       &block_type);
+    if (!blk) return false;
+
+    // calculate word number in block and bit
+    unsigned nbit   = unsigned(n & bm::set_block_mask); 
+
+    if (block_type == 1) // gap
+    {
+        bm::gap_word_t* gap_blk = BMGAP_PTR(blk);
+        bool old_val = (gap_test(gap_blk, nbit) != 0);
+
+        bool new_val = val & old_val;
+        if (new_val != old_val)
+        {
+            unsigned is_set;
+            unsigned new_block_len = 
+                gap_set_value(new_val, gap_blk, nbit, &is_set);
+            BM_ASSERT(is_set);
+            BMCOUNT_ADJ(val)
+
+            unsigned threshold = 
+                bm::gap_limit(gap_blk, blockman_.glen());
+            if (new_block_len > threshold) 
+            {
+                extend_gap_block(nblock, gap_blk);
+            }
+            return true;
+        }
+    }
+    else  // bit block
+    {
+        unsigned nword  = unsigned(nbit >> bm::set_word_shift); 
+        nbit &= bm::set_word_mask;
+
+        bm::word_t* word = blk + nword;
+        bm::word_t  mask = (((bm::word_t)1) << nbit);
+        bool is_set = ((*word) & mask) != 0;
+
+        bool new_val = is_set & val;
+        if (new_val != val)    // need to change bit
+        {
+            if (new_val)       // set bit
+            {
+                *word |= mask;
+                BMCOUNT_INC;
+            }
+            else               // clear bit
+            {
+                *word &= ~mask;
+                BMCOUNT_DEC;
+            }
+            return true;
+        }
+    }
+    return false;
+}
+
+
 // -----------------------------------------------------------------------
 
 template<class Alloc, class MS> 
@@ -2198,7 +2417,9 @@ void bvector<Alloc, MS>::stat(unsigned blocks) const
                int mem_eff = raw_size - cmr_len;
                total_gap_eff += mem_eff;
                
-               printf("[ GAP %i=%i:%i-%i(%i) ]", nb, bc, level, len, mem_eff);
+               unsigned i,j;
+               blockman_.get_block_coord(nb, &i, &j);
+               printf("[ GAP %i(%i,%i)=%i:%i-%i(%i) ]", nb, i, j, bc, level, len, mem_eff);
                 ++printed;
             }
             else // bitset
@@ -2403,7 +2624,7 @@ void bvector<Alloc, MS>::combine_operation(
 
     if (size_ == bvect.size_) 
     {
-        BM_ASSERT( top_blocks >= bvect_top_blocks);
+        BM_ASSERT(top_blocks >= bvect_top_blocks);
     }
     else
     if (size_ < bvect.size_) // this vect shorter than the arg.
@@ -2426,24 +2647,6 @@ void bvector<Alloc, MS>::combine_operation(
             }
         }
     }
-
-    block_bit_op      bit_func;
-    switch (opcode)
-    {
-    case BM_AND:
-        bit_func = bit_block_and;
-        break;
-    case BM_OR:
-        bit_func = bit_block_or;
-        break;
-    case BM_SUB:
-        bit_func = bit_block_sub;
-        break;
-    case BM_XOR:
-        bit_func = bit_block_xor;
-        break;
-    }           
-    
     
     bm::word_t*** blk_root = blockman_.blocks_root();
     unsigned block_idx = 0;
@@ -2451,7 +2654,7 @@ void bvector<Alloc, MS>::combine_operation(
 
     BM_SET_MMX_GUARD
 
-    for (i = 0; i < blockman_.top_block_size(); ++i)
+    for (i = 0; i < top_blocks; ++i)
     {
         bm::word_t** blk_blk = blk_root[i];
 
@@ -2494,41 +2697,6 @@ void bvector<Alloc, MS>::combine_operation(
                 bool arg_gap = BM_IS_GAP(bvect.blockman_, arg_blk, block_idx);
                 bool gap = BM_IS_GAP((*this).blockman_, blk, block_idx);
                 
-                // Optimization branch. Statistically two bit blocks
-                // are bumping into each other quite frequently and
-                // this peace of code tend to be executed often and 
-                // program does not go into nested calls.
-                // But logically this branch can be eliminated without
-                // loosing any functionality.
-/*                   
-                if (!gap && !arg_gap)
-                {
-                    if (IS_VALID_ADDR(blk) && arg_blk)
-                    {
-                    
-                        if (bit_func2 && (j < bm::set_array_size-1))
-                        {
-                            bm::word_t* blk2 = blk_blk[j+1];
-                            const bm::word_t* arg_blk2 = bvect.get_block(i, j+1);
-                            
-                            bool arg_gap2 = BM_IS_GAP(bvect, arg_blk2, block_idx + 1);
-                            bool gap2 = BM_IS_GAP((*this), blk2, block_idx + 1);
-                            
-                            if (!gap2 && !arg_gap2 && blk2 && arg_blk2)
-                            {
-                                bit_func2(blk, arg_blk, blk2, arg_blk2);
-                                ++j;
-                                ++block_idx;
-                                continue;
-                            }
-                            
-                        }
-                        
-                        bit_func(blk, arg_blk);
-                        continue;
-                    }
-                } // end of optimization branch...
-*/                   
                 combine_operation_with_block(block_idx, gap, blk, 
                                             arg_blk, arg_gap,
                                             opcode);
@@ -2565,9 +2733,16 @@ void bvector<Alloc, MS>::combine_operation_with_block(unsigned nb,
         {
             if (arg_gap)  // both blocks GAP-type
             {
-                gap_word_t tmp_buf[bm::gap_max_buff_len * 3]; // temporary result
-            
+                gap_word_t tmp_buf[bm::gap_max_buff_len * 3]; // temporary result            
                 gap_word_t* res;
+
+                gap_operation_func_type gfunc = 
+                    operation_functions<true>::gap_operation(opcode);
+                BM_ASSERT(gfunc);
+                res = (*gfunc)(BMGAP_PTR(blk), 
+                               BMGAP_PTR(arg_blk), 
+                               tmp_buf);
+/*
                 switch (opcode)
                 {
                 case BM_AND:
@@ -2594,11 +2769,11 @@ void bvector<Alloc, MS>::combine_operation_with_block(unsigned nb,
                     assert(0);
                     res = 0;
                 }
-
-                assert(res == tmp_buf);
+*/
+                BM_ASSERT(res == tmp_buf);
                 unsigned res_len = bm::gap_length(res);
 
-                assert(!(res == tmp_buf && res_len == 0));
+                BM_ASSERT(!(res == tmp_buf && res_len == 0));
 
                 // if as a result of the operation gap block turned to zero
                 // we can now replace it with NULL
@@ -2653,22 +2828,22 @@ void bvector<Alloc, MS>::combine_operation_with_block(unsigned nb,
                 
                 if (arg_blk == 0)  // Combining against an empty block
                 {
-                if (opcode == BM_OR  || 
-                    opcode == BM_SUB || 
-                    opcode == BM_XOR)
-                {
-                    return; // nothing to do
-                }
-                    
-                if (opcode == BM_AND) // ("Value" AND  0) == 0
-                {
-                    blockman_.set_block_ptr(nb, 0);
-                    blockman_.set_block_bit(nb);
-                    blockman_.get_allocator().
-                                    free_gap_block(BMGAP_PTR(blk),
-                                                blockman_.glen());
-                    return;
-                }
+                    if (opcode == BM_OR  || 
+                        opcode == BM_SUB || 
+                        opcode == BM_XOR)
+                    {
+                        return; // nothing to do
+                    }
+                        
+                    if (opcode == BM_AND) // ("Value" AND  0) == 0
+                    {
+                        blockman_.set_block_ptr(nb, 0);
+                        blockman_.set_block_bit(nb);
+                        blockman_.get_allocator().
+                                        free_gap_block(BMGAP_PTR(blk),
+                                                    blockman_.glen());
+                        return;
+                    }
                 }
 
                 blk = blockman_.convert_gap2bitset(nb, BMGAP_PTR(blk));
@@ -2678,26 +2853,32 @@ void bvector<Alloc, MS>::combine_operation_with_block(unsigned nb,
         {
             if (arg_gap) // argument block is GAP-type
             {
-            if (IS_VALID_ADDR(blk))
-            {
-                // special case, maybe we can do the job without 
-                // converting the GAP argument to bitblock
-                switch (opcode)
+                if (IS_VALID_ADDR(blk))
                 {
-                case BM_OR:
-                        gap_add_to_bitset(blk, BMGAP_PTR(arg_blk));
-                        return;                         
-                case BM_SUB:
-                        gap_sub_to_bitset(blk, BMGAP_PTR(arg_blk));
-                        return;
-                case BM_XOR:
-                        gap_xor_to_bitset(blk, BMGAP_PTR(arg_blk));
-                        return;
-                case BM_AND:
-                        gap_and_to_bitset(blk, BMGAP_PTR(arg_blk));
-                        return;
-                        
-                } // switch
+                    // special case, maybe we can do the job without 
+                    // converting the GAP argument to bitblock
+                    gap_operation_to_bitset_func_type gfunc = 
+                        operation_functions<true>::gap_op_to_bit(opcode);
+                    BM_ASSERT(gfunc);
+                    (*gfunc)(blk, BMGAP_PTR(arg_blk));
+                    return;
+/*
+                    switch (opcode)
+                    {
+                    case BM_OR:
+                            gap_add_to_bitset(blk, BMGAP_PTR(arg_blk));
+                            return;                         
+                    case BM_SUB:
+                            gap_sub_to_bitset(blk, BMGAP_PTR(arg_blk));
+                            return;
+                    case BM_XOR:
+                            gap_xor_to_bitset(blk, BMGAP_PTR(arg_blk));
+                            return;
+                    case BM_AND:
+                            gap_and_to_bitset(blk, BMGAP_PTR(arg_blk));
+                            return;                            
+                    } // switch
+*/
                 }
                 
                 // the worst case we need to convert argument block to 
@@ -2799,7 +2980,7 @@ void bvector<Alloc, MS>::combine_operation_with_block(unsigned nb,
             }
             break;
         default:
-            assert(0);
+            BM_ASSERT(0);
             ret = 0;
         }
 
@@ -2892,23 +3073,20 @@ void bvector<Alloc, MS>::set_range_no_check(bm::id_t left,
             block = blockman_.get_block(nb);
             if (block == 0)  // nothing to do
                 continue;
-
-            blockman_.set_block(nb, 0);
-            blockman_.set_block_bit(nb);
-
             bool is_gap = BM_IS_GAP(blockman_, block, nb);
+            blockman_.set_block(nb, 0, false /*bit*/);
+            //blockman_.set_block_bit(nb);
 
             if (is_gap) 
             {
                 blockman_.get_allocator().free_gap_block(BMGAP_PTR(block),
-                                                            blockman_.glen());
+                                                         blockman_.glen());
             }
             else
             {
                 blockman_.get_allocator().free_bit_block(block);
             }
 
-
         } // for
     } // if value else 
 
@@ -2937,6 +3115,11 @@ void bvector<Alloc, MS>::set_range_no_check(bm::id_t left,
 
 #include "bmundef.h"
 
+#ifdef _MSC_VER
+#pragma warning( default : 4311 4312)
+#endif
+
+
 #endif
 
 
