@@ -1,8 +1,7 @@
 /*
 Copyright(c) 2002-2005 Anatoliy Kuznetsov(anatoliy_kuznetsov at yahoo.com)
 
-PermiFor more information please visit:  http://bmagic.sourceforge.net
-ssion is hereby granted, free of charge, to any person 
+Permission is hereby granted, free of charge, to any person 
 obtaining a copy of this software and associated documentation 
 files (the "Software"), to deal in the Software without restriction, 
 including without limitation the rights to use, copy, modify, merge, 
@@ -34,6 +33,7 @@ For more information please visit:  http://bmagic.sourceforge.net
 namespace bm
 {
 
+
 /*!
    @brief bitvector blocks manager
         Embedded class managing bit-blocks on very low level.
@@ -49,12 +49,14 @@ public:
 
     typedef Alloc allocator_type;
 
-    /** Base functor class */
+    /** Base functor class (block visitor)*/
     class bm_func_base
     {
     public:
         bm_func_base(blocks_manager& bman) : bm_(bman) {}
 
+        void on_empty_top(unsigned /* top_block_idx*/ ) {}
+        void on_empty_block(unsigned /* block_idx*/ ) {}
     protected:
         blocks_manager&  bm_;
     };
@@ -66,6 +68,8 @@ public:
     public:
         bm_func_base_const(const blocks_manager& bman) : bm_(bman) {}
 
+        void on_empty_top(unsigned /* top_block_idx*/ ) {}
+        void on_empty_block(unsigned /* block_idx*/ ) {}
     protected:
         const blocks_manager&  bm_;
     };
@@ -80,25 +84,7 @@ public:
 
         bm::id_t block_count(const bm::word_t* block, unsigned idx) const
         {
-            id_t count = 0;
-            if (IS_FULL_BLOCK(block))
-            {
-                count = bm::bits_in_block;
-            }
-            else
-            {
-                if (BM_IS_GAP(this->bm_, block, idx))
-                {
-                    count = gap_bit_count(BMGAP_PTR(block));
-                }
-                else // bitset
-                {
-                    count = 
-                        bit_block_calc_count(block, 
-                                                block + bm::set_block_size);
-                }
-            }
-            return count;
+            return this->bm_.block_bitcount(block, idx);
         }
     };
 
@@ -176,7 +162,7 @@ public:
             }
             else
             {
-                if (BM_IS_GAP(this->bm_, block, idx))
+                if (BM_IS_GAP(*this, block, idx))
                 {
                     gap_word_t* gap_block = BMGAP_PTR(block);
                     count = gap_length(gap_block) - 1;
@@ -231,7 +217,7 @@ public:
         {
             if (IS_FULL_BLOCK(block)) return true;
 
-            if (BM_IS_GAP(this->bm_, block, idx)) // gap block
+            if (BM_IS_GAP(*this, block, idx)) // gap block
             {
                 if (!gap_is_all_zero(BMGAP_PTR(block), bm::gap_max_bits))
                 {
@@ -267,7 +253,7 @@ public:
         {
             blocks_manager& bman = this->bm_;
             
-            if (!BM_IS_GAP(bman, block, idx))
+            if (!BM_IS_GAP(*this, block, idx))
                 return;
 
             gap_word_t* gap_blk = BMGAP_PTR(block);
@@ -322,38 +308,75 @@ public:
     public:
         block_opt_func(blocks_manager& bm, 
                         bm::word_t*     temp_block,
-                        int             opt_mode) 
+                        int             opt_mode,
+                        bv_statistics*  bv_stat=0) 
             : bm_func_base(bm),
-                temp_block_(temp_block),
-                opt_mode_(opt_mode)
+              temp_block_(temp_block),
+              opt_mode_(opt_mode),
+              stat_(bv_stat),
+              empty_(0)
         {
             BM_ASSERT(temp_block);
         }
 
+        void on_empty_top(unsigned i)
+        {
+            bm::word_t*** blk_root = this->bm_.get_rootblock();
+            bm::word_t** blk_blk = blk_root[i];
+            if (blk_blk) 
+            {
+                this->bm_.get_allocator().free_ptr(blk_blk);
+                blk_root[i] = 0;
+            }
+            if (stat_)
+            {
+                stat_->max_serialize_mem += sizeof(unsigned) + 1;
+            }
+        }
+        void on_empty_block(unsigned /* block_idx*/ ) { ++empty_; }
+
         void operator()(bm::word_t* block, unsigned idx)
         {
             blocks_manager& bman = this->bm_;
-            if (IS_FULL_BLOCK(block)) return;
+            if (IS_FULL_BLOCK(block)) 
+            {
+                ++empty_;
+                return;
+            }
+
+            if (stat_) 
+            {
+                stat_->max_serialize_mem += empty_ << 2;
+                empty_ = 0;
+            }
 
             gap_word_t* gap_blk;
-
-            if (BM_IS_GAP(bman, block, idx)) // gap block
+            if (BM_IS_GAP(*this, block, idx)) // gap block
             {
                 gap_blk = BMGAP_PTR(block);
-
+                // check if block is empty (or all 1)
                 if (gap_is_all_zero(gap_blk, bm::gap_max_bits))
                 {
                     bman.set_block_ptr(idx, 0);
-                    goto free_block;
+                    this->free_block(gap_blk, idx);
+                    ++empty_;
                 }
                 else 
                 if (gap_is_all_one(gap_blk, bm::gap_max_bits))
                 {
                     bman.set_block_ptr(idx, FULL_BLOCK_ADDR);
-                free_block:
-                    bman.get_allocator().free_gap_block(gap_blk, 
-                                                        bman.glen());
-                    bman.set_block_bit(idx);
+                    this->free_block(gap_blk, idx);
+                    ++empty_;
+                }
+                else
+                {
+                    // regular gap block - compute statistics
+                    if (stat_)
+                    {
+                        stat_->add_gap_block(
+                                    bm::gap_capacity(gap_blk, bman.glen()),
+                                    bm::gap_length(gap_blk));
+                    }
                 }
             }
             else // bit block
@@ -369,8 +392,9 @@ public:
                     {
                         bman.get_allocator().free_bit_block(block);
                         bman.set_block_ptr(idx, 0);
-                        return;
-                    }
+                        ++empty_;
+                    } 
+                    else
                     if (opt_mode_ == 2) // check if it is all 1 block
                     {
                         b = is_bits_one(blk1, blk2);
@@ -378,9 +402,13 @@ public:
                         {
                             bman.get_allocator().free_bit_block(block);
                             bman.set_block_ptr(idx, FULL_BLOCK_ADDR);
-                            return;
+                            ++empty_;
                         }
                     }
+
+                    if (!b && stat_)
+                        stat_->add_bit_block();
+
                     return;
                 }
             
@@ -395,41 +423,57 @@ public:
                                                     block, 
                                                     bm::gap_max_bits, 
                                                     threashold);
+                if (len)    // compression successful                
+                {                
+                    bman.get_allocator().free_bit_block(block);
 
+                    // check if new gap block can be eliminated
+                    if (gap_is_all_zero(tmp_gap_blk, bm::gap_max_bits))
+                    {
+                        bman.set_block_ptr(idx, 0);
+                        ++empty_;
+                    }
+                    else if (gap_is_all_one(tmp_gap_blk, bm::gap_max_bits))
+                    {
+                        bman.set_block_ptr(idx, FULL_BLOCK_ADDR);
+                        ++empty_;
+                    }
+                    else
+                    {
+                        int level = bm::gap_calc_level(len, bman.glen());
 
-                if (!len) return;
-                
-                // convertion successful
-                
-                bman.get_allocator().free_bit_block(block);
-
-                // check if new gap block can be eliminated
-
-                if (gap_is_all_zero(tmp_gap_blk, bm::gap_max_bits))
+                        gap_blk = 
+                            bman.allocate_gap_block(level, tmp_gap_blk);
+                        bman.set_block_ptr(idx, (bm::word_t*)gap_blk);
+                        bman.set_block_gap(idx);
+                        if (stat_)
+                        {
+                            stat_->add_gap_block(
+                                    bm::gap_capacity(gap_blk, bman.glen()),
+                                    bm::gap_length(gap_blk));
+                        }
+                    }
+                }  
+                else  // non-compressable bit-block
                 {
-                    bman.set_block_ptr(idx, 0);
+                    if (stat_)
+                        stat_->add_bit_block();
                 }
-                else if (gap_is_all_one(tmp_gap_blk, bm::gap_max_bits))
-                {
-                    bman.set_block_ptr(idx, FULL_BLOCK_ADDR);
-                }
-                else
-                {
-                    int level = bm::gap_calc_level(len, bman.glen());
-
-                    gap_blk = 
-                        bman.allocate_gap_block(level, tmp_gap_blk);
-                    bman.set_block_ptr(idx, (bm::word_t*)gap_blk);
-                    bman.set_block_gap(idx);
-                }
-                
-        
             }
-
         }
     private:
-        bm::word_t*   temp_block_;
-        int           opt_mode_;
+        void free_block(gap_word_t* gap_blk, unsigned idx)
+        {
+            this->bm_.get_allocator().free_gap_block(gap_blk,
+                                                     this->bm_.glen());
+            this->bm_.set_block_bit(idx);
+        }
+
+    private:
+        bm::word_t*         temp_block_;
+        int                 opt_mode_;
+        bv_statistics*      stat_;
+        unsigned            empty_;
     };
 
     /** Bitblock invert functor */
@@ -452,7 +496,7 @@ public:
             }
             else
             {
-                if (BM_IS_GAP(this->bm_, block, idx)) // gap block
+                if (BM_IS_GAP(*this, block, idx)) // gap block
                 {
                     gap_invert(BMGAP_PTR(block));
                 }
@@ -486,7 +530,7 @@ public:
             }
             else
             {
-                if (BM_IS_GAP(bman, block, idx))
+                if (BM_IS_GAP(*this, block, idx))
                 {
                     gap_set_all(BMGAP_PTR(block), bm::gap_max_bits, 0);
                 }
@@ -556,12 +600,12 @@ public:
 
         void operator()(bm::word_t* block, unsigned idx)
         {
-            bool gap = bm_src_.is_block_gap(idx);
+            bool is_gap = bm_src_.is_block_gap(idx);
             bm::word_t* new_blk;
             
             blocks_manager& bman = this->bm_;
 
-            if (gap)
+            if (is_gap)
             {
                 bm::gap_word_t* gap_block = BMGAP_PTR(block); 
                 unsigned level = gap_level(gap_block);
@@ -570,7 +614,7 @@ public:
                                                         bman.glen());
                 int len = gap_length(BMGAP_PTR(block));
                 ::memcpy(new_blk, gap_block, len * sizeof(gap_word_t));
-                BMSET_PTRGAP(new_blk);
+                //BMSET_PTRGAP(new_blk);
             }
             else
             {
@@ -584,7 +628,7 @@ public:
                     bit_block_copy(new_blk, block);
                 }
             }
-            bman.set_block(idx, new_blk);
+            bman.set_block(idx, new_blk, is_gap);
         }
 
     private:
@@ -617,11 +661,13 @@ public:
         volatile const char* vp = _copyright<true>::_p;
         char c = *vp;
         c = 0;
+        effective_top_block_size_ = 1;
     }
 
     blocks_manager(const blocks_manager& blockman)
         : blocks_(0),
             top_block_size_(blockman.top_block_size_),
+            effective_top_block_size_(blockman.effective_top_block_size_),
         #ifdef BM_DISBALE_BIT_IN_PTR
             gap_flags_(blockman.gap_flags_),
         #endif
@@ -721,30 +767,40 @@ public:
         return blk_blk ? blk_blk[nb & bm::set_array_mask] : 0;
     }
 
+    /** 
+    Recalculate absolute block address into coordinates
+    */
+    void get_block_coord(unsigned nb, unsigned* i, unsigned* j) const
+    {
+        BM_ASSERT(i);
+        BM_ASSERT(j);
+
+        *i = nb >> bm::set_array_shift; // top block address
+        *j = nb &  bm::set_array_mask;  // address in sub-block
+    }
+
     bool is_no_more_blocks(unsigned nb) const
     {
-        unsigned block = nb;
-        unsigned block_idx = nb >> bm::set_array_shift;
-        for (unsigned i = block_idx; i < top_block_size_; ++i) { 
+        unsigned i,j;
+        get_block_coord(nb, &i, &j);
+        for (;i < effective_top_block_size_; ++i) 
+        { 
             bm::word_t** blk_blk = blocks_[i];
-            if (blk_blk)                 
+            if (!blk_blk)
             { 
-                for (unsigned j = block & bm::set_array_mask; 
-                        j < bm::set_array_size; ++j)
-                {
-                    bm::word_t* blk = blk_blk[j];
-                    if (blk && !is_block_zero(i, blk)) 
-                    {
-                        return false;
-                    }
-                    ++block;
-                }
-            } 
-            else 
-            {
-                block += bm::set_array_size;
+                nb += bm::set_array_size;
             }
-        }
+            else
+               for (;j < bm::set_array_size; ++j, ++nb)
+               {
+                   bm::word_t* blk = blk_blk[j];
+                   if (blk && !is_block_zero(nb, blk)) 
+                   {
+                       return false;
+                   }
+               } // for j
+            j = 0;
+        } // for i
         return true;
     }
 
@@ -792,7 +848,7 @@ public:
         set_block_bit(nb);    
         #endif
 
-        if (BM_IS_GAP((*this), block, nb))
+        if (BM_IS_GAP(*this, block, nb))
         {
             alloc_.free_gap_block(BMGAP_PTR(block), glevel_len_);
         }
@@ -802,6 +858,28 @@ public:
         }
     }
 
+    /**
+        Create all-zeros bit block. Old block (if exists) is deleted.
+    */
+    bm::word_t* make_bit_block(unsigned nb)
+    {
+        bm::word_t* block = this->get_allocator().alloc_bit_block();
+        bit_block_set(block, 0);
+        bm::word_t* old_block = set_block(nb, block);
+        if (IS_VALID_ADDR(old_block))
+        {
+            if (BM_IS_GAP(*this, old_block, nb))
+            {
+                alloc_.free_gap_block(BMGAP_PTR(old_block), glen());
+            }
+            else
+            {
+                alloc_.free_bit_block(old_block);
+            }
+        }
+        return block;
+    }
+
     /** 
         Function checks if block is not yet allocated, allocates it and sets to
         all-zero or all-one bits. 
@@ -832,24 +910,17 @@ public:
             if (initial_block_type == 0) // bitset requested
             {
                 block = alloc_.alloc_bit_block();
-
                 // initialize block depending on its previous status
-
                 bit_block_set(block, block_flag ? 0xFF : 0);
-
                 set_block(nb, block);
             }
             else // gap block requested
             {
                 bm::gap_word_t* gap_block = allocate_gap_block(0);
                 gap_set_all(gap_block, bm::gap_max_bits, block_flag);
-                set_block(nb, (bm::word_t*)gap_block);
-
-                set_block_gap(nb);
-
+                set_block(nb, (bm::word_t*)gap_block, true/*gap*/);
                 return (bm::word_t*)gap_block;
             }
-
         }
         else // block already exists
         {
@@ -892,6 +963,67 @@ public:
         // auto-resize the top block array
         if (nblk_blk >= top_block_size_)
             reserve_top_blocks(nblk_blk + 1);
+        if (nblk_blk >= effective_top_block_size_)
+            effective_top_block_size_ = nblk_blk + 1;
+
+        // If first level array not yet allocated, allocate it and
+        // assign block to it
+        if (blocks_[nblk_blk] == 0) 
+        {
+            blocks_[nblk_blk] = (bm::word_t**)alloc_.alloc_ptr();
+            ::memset(blocks_[nblk_blk], 0, 
+                bm::set_array_size * sizeof(bm::word_t*));
+
+            old_block = 0;
+        }
+        else
+        {
+            old_block = blocks_[nblk_blk][nb & bm::set_array_mask];
+        }
+
+        // NOTE: block will be replaced without freeing,
+        // potential memory leak may lay here....
+        blocks_[nblk_blk][nb & bm::set_array_mask] = block; // equivalent to %
+
+        return old_block;
+    }
+
+    /**
+        Places new block into descriptors table, returns old block's address.
+        Old block is not deleted.
+    */
+    bm::word_t* set_block(unsigned nb, bm::word_t* block, bool gap)
+    {
+        bm::word_t* old_block;
+
+        if (block)
+        {
+            if (gap)
+            {
+            #ifdef BM_DISBALE_BIT_IN_PTR
+                gap_flags_.set(nb);
+            #else
+                block = (bm::word_t*)BMPTR_SETBIT0(block);
+            #endif
+            }
+            else 
+            {
+            #ifdef BM_DISBALE_BIT_IN_PTR
+                gap_flags_.set(nb, false);
+            #else
+                block = (bm::word_t*)BMPTR_CLEARBIT0(block);
+            #endif
+            }
+        }
+
+        // top block index
+        register unsigned nblk_blk = nb >> bm::set_array_shift;
+
+        // auto-resize the top block array
+        if (nblk_blk >= top_block_size_)
+            reserve_top_blocks(nblk_blk + 1);
+        if (nblk_blk >= effective_top_block_size_)
+            effective_top_block_size_ = nblk_blk + 1;
 
         // If first level array not yet allocated, allocate it and
         // assign block to it
@@ -920,6 +1052,7 @@ public:
     */
     void set_block_ptr(unsigned nb, bm::word_t* block)
     {
+        BM_ASSERT((nb >> bm::set_array_shift) < effective_top_block_size_);
         blocks_[nb >> bm::set_array_shift][nb & bm::set_array_mask] = block;
     }
         
@@ -928,7 +1061,7 @@ public:
         \param nb - Block's index. 
         \param gap_block - Pointer to the gap block, 
                             if NULL block nb will be taken
-        \return new gap block's memory
+        \return new bit block's memory
     */
     bm::word_t* convert_gap2bitset(unsigned nb, gap_word_t* gap_block=0)
     {
@@ -958,6 +1091,71 @@ public:
         return new_block;
     }
 
+    /**
+        Make sure block turns into true bit-block if it is GAP or a full block
+    */
+    bm::word_t* deoptimize_block(unsigned nb)
+    {
+        bm::word_t* block = this->get_block(nb);
+        if (BM_IS_GAP(*this, block, nb))
+        {
+            return convert_gap2bitset(nb);
+        }
+        if (IS_FULL_BLOCK(block)) 
+        {
+            bm::word_t* new_block = get_allocator().alloc_bit_block();
+            bit_block_copy(new_block, block);
+            set_block(nb, new_block);
+            return new_block;
+        }
+        return block;
+    }
+
+    /**
+    Free block, make it zero pointer in the tree
+    */
+    bm::word_t* zero_block(unsigned nb)
+    {
+        bm::word_t* block = this->get_block(nb);
+        if (!block) return block;
+        if (BM_IS_GAP(*this, block, nb)) // gap block
+        {
+            get_allocator().free_gap_block(BMGAP_PTR(block), glen());
+        }
+        else
+        {
+            // deallocates only valid pointers
+            get_allocator().free_bit_block(block);
+        }
+        set_block(nb, 0);
+        return 0;
+    }
+
+    /**
+        Count number of bits ON in the block
+    */
+    bm::id_t block_bitcount(const bm::word_t* block, unsigned idx) const
+    { 
+        id_t count = 0;
+        if (!block) return count;
+        if (IS_FULL_BLOCK(block))
+            count = bm::bits_in_block;
+        else
+        {
+            if (BM_IS_GAP(*this, block, idx))
+            {
+                count = gap_bit_count(BMGAP_PTR(block));
+            }
+            else // bitset
+            {
+                count = 
+                    bit_block_calc_count(block, 
+                                         block + bm::set_block_size);
+            }
+        }
+        return count;
+    }
+
     /**
         \brief Extends GAP block to the next level or converts it to bit block.
         \param nb - Block's linear index.
@@ -1025,7 +1223,7 @@ public:
     {
         if (blk == 0) return true;
 
-        if (BM_IS_GAP((*this), blk, nb)) // GAP
+        if (BM_IS_GAP(*this, blk, nb)) // GAP
         {
             gap_word_t* b = BMGAP_PTR(blk);
             return gap_is_all_zero(b, bm::gap_max_bits);
@@ -1050,7 +1248,7 @@ public:
     {
         if (blk == 0) return false;
 
-        if (BM_IS_GAP((*this), blk, nb)) // GAP
+        if (BM_IS_GAP(*this, blk, nb)) // GAP
         {
             gap_word_t* b = BMGAP_PTR(blk);
             return gap_is_all_one(b, bm::gap_max_bits);
@@ -1163,6 +1361,7 @@ public:
         #endif
 
 		xor_swap(this->top_block_size_, bm.top_block_size_);
+        xor_swap(this->effective_top_block_size_, bm.effective_top_block_size_);
 
         BM_ASSERT(sizeof(glevel_len_) / sizeof(glevel_len_[0]) 
                                     == bm::gap_levels); // paranoiya check
@@ -1179,6 +1378,33 @@ public:
         return top_block_size_;
     }
 
+    /*! \brief Returns effective size of the top block array in the tree 
+    Effective size excludes NULL pointers at the top descriptor end
+    */
+    unsigned effective_top_block_size() const
+    {
+        return effective_top_block_size_;
+/*
+        unsigned i = top_block_size();
+        if (!i) return 0;
+        for (--i; i > 0; --i) 
+        {
+            if (blocks_[i] != 0) 
+            {
+                if (this->effective_top_block_size_ < i+1)
+                {
+printf("Effective size error!\n");
+                exit(1);
+                }
+                return i+1;
+            }
+        }
+        BM_ASSERT(this->effective_top_block_size_ >= 1);
+        return 1;
+*/        
+
+    }
+
     /**
         \brief reserve capacity for specified number of bits
     */
@@ -1229,12 +1455,12 @@ private:
     void deinit_tree()
     {
         if (blocks_ == 0) return;
-
+        unsigned top_size = this->effective_top_block_size();
         block_free_func  free_func(*this);
-        for_each_nzblock(blocks_, top_block_size_, 
-                                    bm::set_array_size, free_func);
-                                    
-        for(unsigned i = 0; i <  top_block_size_; ++i)
+        for_each_nzblock(blocks_, top_size, 
+                                  bm::set_array_size, free_func);
+
+        for(unsigned i = 0; i < top_size; ++i)
         {
             bm::word_t** blk_blk = blocks_[i];
             if (blk_blk) 
@@ -1248,6 +1474,8 @@ private:
     bm::word_t***                          blocks_;
     /// Size of the top level block array in blocks_ tree
     unsigned                               top_block_size_;
+    /// Effective size of the top level block array in blocks_ tree
+    unsigned                               effective_top_block_size_;
     #ifdef BM_DISBALE_BIT_IN_PTR
     /// mini bitvector used to indicate gap blocks
     MS                                     gap_flags_;
@@ -1260,6 +1488,38 @@ private:
     allocator_type                         alloc_;
 };
 
+/**
+    Bit block buffer guard
+    \internal
+*/
+template<class BlocksManager>
+class bit_block_guard
+{
+public:
+    bit_block_guard(BlocksManager& bman, bm::word_t* blk=0) 
+        : bman_(bman), 
+          block_(blk)
+    {}
+    ~bit_block_guard()
+    {
+        bman_.get_allocator().free_bit_block(block_);
+    }
+    void attach(bm::word_t* blk)
+    {
+        bman_.get_allocator().free_bit_block(block_);
+        block_ = blk;
+    }
+    bm::word_t* allocate()
+    {
+        attach(bman_.get_allocator().alloc_bit_block());
+        return block_;
+    }
+    bm::word_t* get() { return block_; }
+private:
+    BlocksManager& bman_;
+    bm::word_t*    block_;
+};
+
 
 }
 
